/**
 * @license
 * Copyright 2022 Dynatrace LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { Blob } from 'buffer';
import { magenta, red, yellow } from 'chalk';
import * as Console from 'console';
import { webcrypto } from 'crypto';
import { cpus } from 'os';
import {
  ByteLengthQueuingStrategy,
  CountQueuingStrategy,
  ReadableByteStreamController,
  ReadableStream,
  ReadableStreamBYOBReader,
  ReadableStreamBYOBRequest,
  ReadableStreamDefaultController,
  ReadableStreamDefaultReader,
  TextDecoderStream,
  TextEncoderStream,
  TransformStream,
  TransformStreamDefaultController,
  WritableStream,
  WritableStreamDefaultController,
  WritableStreamDefaultWriter,
} from 'stream/web';
import { File, FileReader, FormData, Headers, Request, Response } from 'undici';
import { URLPattern } from 'urlpattern-polyfill';
import {
  performance,
  PerformanceEntry,
  PerformanceMark,
  PerformanceMeasure,
} from 'perf_hooks';
import { TextDecoder, TextEncoder } from 'util';
import { deserialize, serialize } from 'v8';
import * as vm from 'vm';
import atob from './polyfills/atob';
import btoa from './polyfills/btoa';
import fetch from './polyfills/fetch';
import {
  CustomEvent,
  ErrorEvent,
  ProgressEvent,
  PromiseRejectionEvent,
} from './polyfills/events';
import DOMException from './polyfills/domexception';
import Location from './polyfills/location';
import Navigator from './polyfills/navigator';
import Window from './polyfills/window';

export interface ContextConfiguration {
  appId: string;
  appName: string;
  environmentUrl: string;
  token: string;
  tenantId?: string;
}

export interface RuntimeContext {
  /** The Node.js `vm` context */
  vmContext: vm.Context;
  /**
   * This function should be called from outside when an unhandled promise rejection was detected
   * @returns `false` if `preventDefault` was called, otherwise `true`
   */
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  reportUnhandledRejection: (reason: any, promise: Promise<unknown>) => boolean;
}

/** Creates a new VM context */
export function createContext(
  cwd: string,
  urlPathname: string,
  config?: ContextConfiguration,
): RuntimeContext {
  const environmentUrl = config?.environmentUrl ?? 'http://127.0.0.1';

  const globals: PropertyDescriptorMap = {
    environmentUrl: readOnly(environmentUrl),
  };

  if (config) {
    if (Object.prototype.hasOwnProperty.call(config, 'appId')) {
      globals.appId = readOnly(config.appId);
    }

    if (Object.prototype.hasOwnProperty.call(config, 'appName')) {
      globals.appName = readOnly(config.appName);
    }

    if (Object.prototype.hasOwnProperty.call(config, 'tenantId')) {
      globals.tenantId = readOnly(config.tenantId);
    }
  }

  const href = new URL(environmentUrl);
  href.pathname = urlPathname!;

  const location = new Location(href.toString());

  const navigator = Object.create(Navigator);
  Object.defineProperty(navigator, 'hardwareConcurrency', {
    get(): number {
      return cpus().length;
    },
    configurable: true,
    enumerable: true,
  });

  const globalProperties = Object.defineProperties(new Window(), {
    ...globals,
    AbortController: nonEnumerable(AbortController),
    AbortSignal: nonEnumerable(AbortSignal),
    Blob: nonEnumerable(Blob),
    ByteLengthQueuingStrategy: nonEnumerable(ByteLengthQueuingStrategy),
    // TODO: This API will become available in Node.js after the version bump to v18 (ADES-2336)
    CompressionStream: nonEnumerable(undefined),
    CountQueuingStrategy: nonEnumerable(CountQueuingStrategy),
    CryptoKey: nonEnumerable(webcrypto.CryptoKey),
    CustomEvent: nonEnumerable(CustomEvent),
    // TODO: This API will become available in Node.js after the version bump to v18 (ADES-2336)
    DecompressionStream: nonEnumerable(undefined),
    DOMException: nonEnumerable(DOMException),
    ErrorEvent: nonEnumerable(ErrorEvent),
    Event: nonEnumerable(Event),
    EventTarget: nonEnumerable(EventTarget),
    File: nonEnumerable(File),
    FileReader: nonEnumerable(FileReader),
    FormData: nonEnumerable(FormData),
    Headers: nonEnumerable(Headers),
    Performance: nonEnumerable(performance.constructor),
    PerformanceEntry: nonEnumerable(PerformanceEntry),
    PerformanceMark: nonEnumerable(PerformanceMark),
    PerformanceMeasure: nonEnumerable(PerformanceMeasure),
    PromiseRejectionEvent: nonEnumerable(PromiseRejectionEvent),
    Location: {
      value: Location,
      configurable: true,
      writable: true,
    },
    location: {
      get(): Location {
        return location;
      },
      set() {
        // eslint-disable-next-line quotes
        throw new DOMException(`Cannot set "location".`, 'NotSupportedError');
      },
      enumerable: true,
    },
    ProgressEvent: nonEnumerable(ProgressEvent),
    ReadableStream: nonEnumerable(ReadableStream),
    ReadableStreamDefaultReader: nonEnumerable(ReadableStreamDefaultReader),
    TextDecoder: nonEnumerable(TextDecoder),
    TextEncoder: nonEnumerable(TextEncoder),
    TextDecoderStream: nonEnumerable(TextDecoderStream),
    TextEncoderStream: nonEnumerable(TextEncoderStream),
    TransformStream: nonEnumerable(TransformStream),
    URL: nonEnumerable(URL),
    // Note: This API isn't available in the TypeScript DOM bindings yet
    // https://github.com/microsoft/TypeScript-DOM-lib-generator/issues/1199
    URLPattern: nonEnumerable(URLPattern),
    URLSearchParams: nonEnumerable(URLSearchParams),
    WritableStream: nonEnumerable(WritableStream),
    WritableStreamDefaultWriter: nonEnumerable(WritableStreamDefaultWriter),
    WritableStreamDefaultController: nonEnumerable(
      WritableStreamDefaultController,
    ),
    ReadableByteStreamController: nonEnumerable(ReadableByteStreamController),
    ReadableStreamBYOBReader: nonEnumerable(ReadableStreamBYOBReader),
    ReadableStreamBYOBRequest: nonEnumerable(ReadableStreamBYOBRequest),
    ReadableStreamDefaultController: nonEnumerable(
      ReadableStreamDefaultController,
    ),
    TransformStreamDefaultController: nonEnumerable(
      TransformStreamDefaultController,
    ),
    Request: nonEnumerable(Request),
    Response: nonEnumerable(Response),
    atob: writable(atob),
    btoa: writable(btoa),
    setInterval: writable(setInterval),
    setTimeout: writable(setTimeout),
    clearInterval: writable(clearInterval),
    clearTimeout: writable(clearTimeout),
    console: nonEnumerable(
      new Proxy(console, {
        get(target: Console, prop: keyof Console) {
          switch (prop) {
            case 'log':
            case 'info':
            case 'debug':
              return (...args) =>
                console[prop].call(
                  console,
                  magenta(`[${prop.toUpperCase()}]`),
                  yellow(urlPathname),
                  ...args,
                );
            case 'warn':
            case 'error':
              // Prevent console.warn and console.error from writing to stderr output
              return (...args) =>
                console.log(
                  red(`[${prop.toUpperCase()}]`),
                  yellow(urlPathname),
                  ...args,
                );
            default:
              return target[prop];
          }
        },
      }),
    ),
    crypto: readOnly(webcrypto),
    Crypto: nonEnumerable(webcrypto.constructor),
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    SubtleCrypto: nonEnumerable(webcrypto.subtle),
    fetch: writable(fetch(cwd, environmentUrl, config)),
    performance: readOnly(performance),
    queueMicrotask: nonEnumerable(queueMicrotask),
    // reportError is implemented below
    /* eslint-disable @typescript-eslint/no-explicit-any */
    structuredClone: nonEnumerable(
      globalThis.structuredClone ??
        function (value: any): any {
          return deserialize(serialize(value));
        },
    ),
    /* eslint-enable */
    Navigator: nonEnumerable(Navigator),
    navigator: {
      get() {
        return navigator;
      },
      configurable: true,
      enumerable: true,
    },
    Window: {
      value: Window,
      configurable: true,
      enumerable: false,
      writable: true,
    },
  });

  const window = Object.setPrototypeOf(
    Object.defineProperties(globalProperties, {
      window: readOnly(globalProperties),
      self: readOnly(globalProperties),
    }),
    Window.prototype,
  );
  // Monkey-patch `reportError` since `dispatchEvent` is only available after assigning
  // the `Window` prototype to the global object
  // Note that this isn't an accurate implementation since we're missing data such as
  // the file name and position, but it's probably close enough for most cases
  window.reportError = (error: Error) => {
    const ev = new ErrorEvent('error', {
      colno: 0,
      error,
      filename: '',
      lineno: 0,
      message: error?.message ?? '',
      cancelable: true,
    });
    if (window.dispatchEvent(ev)) {
      // `queueMicrotask` schedules the code to run sometime later in the event loop.
      // This emulates the expected semantics for `reportError`, where errors can only be handled
      // via a `window` event rather than with `try`/`catch`
      queueMicrotask(() => {
        throw error;
      });
    }
  };

  // This function must be called from outside to report unhandled promise rejections
  const reportUnhandledRejection = (reason, promise) => {
    const ev = new PromiseRejectionEvent('unhandledrejection', {
      reason,
      promise,
      cancelable: true,
    });
    return window.dispatchEvent(ev);
  };

  return {
    vmContext: vm.createContext(window),
    reportUnhandledRejection,
  };
}

/** Writable property descriptor factory */
function writable(
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  value: any,
): PropertyDescriptor {
  return {
    value,
    configurable: true,
    enumerable: true,
    writable: true,
  };
}

/** Non-enumerable property descriptor factory */
function nonEnumerable(
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  value: any,
): PropertyDescriptor {
  return {
    value,
    configurable: true,
    enumerable: false,
    writable: true,
  };
}

/** Read-only property descriptor factory */
function readOnly(
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  value: any,
): PropertyDescriptor {
  return {
    value,
    configurable: true,
    enumerable: true,
    writable: false,
  };
}
