/**
 * @license
 * Copyright 2022 Dynatrace LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import * as cp from 'child_process';
import { createServer } from 'net';
import { EventEmitter } from 'events';
import { Readable, Writable } from 'stream';
import { serialize } from 'v8';
import { dispatch, initialize } from './controller';
import type { Task } from './task';
import { findPortInRange } from '@dynatrace/devkit';

beforeEach(() => {
  jest.spyOn(console, 'log').mockImplementation(() => undefined);
});

test('should connect to the worker server', async () => {
  const port = await findPortInRange('localhost', 8000);
  const mockServer = createServer();
  mockServer.maxConnections = 1;
  mockServer.once('connection', (socket) => {
    socket.once('data', () => {
      socket.write(
        serialize({
          result: {
            foo: 'bar',
            test: [1, 2, 3, 'a', 'b'],
          },
        }),
        (error: Error | undefined) => {
          if (error) {
            throw error;
          }

          socket.end();
        },
      );
    });
  });

  const cpm = new EventEmitter() as cp.ChildProcess;
  cpm.stdin = new Writable();
  cpm.stdout = new Proxy(new EventEmitter(), {
    get(target: EventEmitter, prop: string | symbol) {
      if (prop === 'on') {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        return (event, listener: (chunk: any) => void) => {
          target[prop](event, listener);
          if (event === 'data') {
            target.emit('data', 'Listening on port ' + port);
          }
        };
      }

      return target[prop];
    },
  }) as Readable;
  cpm.stderr = new EventEmitter() as Readable;

  const cmdSpy: jest.SpyInstance<cp.ChildProcess> = jest
    .spyOn(cp, 'spawn')
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    .mockImplementation(() => {
      mockServer.listen(port);

      return cpm;
    });

  const task: Task = {
    cwd: process.cwd(),
    urlPathname: '/api/test',
    scriptName: 'test.js',
    code: Buffer.alloc(0),
    config: {
      appId: 'my.app.id',
      appName: 'my.app.name',
      tenantId: 'my-tenant',
      environmentUrl: 'https://environment.localhost.test',
      token: 'my-token',
    },
  };

  await initialize();
  const result = await dispatch<unknown>(task);

  expect(cmdSpy).toHaveBeenCalled();

  expect(result).toEqual({
    foo: 'bar',
    test: [1, 2, 3, 'a', 'b'],
  });
  expect.assertions(2);

  mockServer.close();
});
