/**
 * @license
 * Copyright 2022 Dynatrace LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { isAsyncFunction } from 'util/types';
import { join } from 'path';
import * as vm from 'vm';
import type { Task } from './task';
import { createContext } from './context';

type ModuleNamespace = {
  default: (unknown) => unknown | unknown;
};

/** Executes the code in a new VM context */
export async function run(task: Task): Promise<unknown> {
  if (!Object.prototype.hasOwnProperty.call(vm, 'SourceTextModule')) {
    throw new Error(
      'Modules support requires the --experimental-vm-modules flag',
    );
  }

  const { vmContext, reportUnhandledRejection } = createContext(
    task.cwd,
    task.urlPathname,
    task.config,
  );
  const context = vm.createContext(vmContext);

  let unhandledPromise: unknown = undefined;
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  const onUnhandledRejection = (reason: any, promise: Promise<unknown>) => {
    if (reportUnhandledRejection(reason, promise)) {
      console.error(`Unhandled promise rejection: ${reason}`);
      unhandledPromise = promise;
    }
  };
  process.on('unhandledRejection', onUnhandledRejection);

  const module = new vm.SourceTextModule(task.code.toString(), {
    identifier: join(task.cwd, task.scriptName),
    context,
  });
  await module.link((async () => undefined) as unknown as vm.ModuleLinker);
  await module.evaluate();

  let result;
  const namespace = module.namespace as ModuleNamespace;
  if (Object.prototype.hasOwnProperty.call(namespace, 'default')) {
    if (
      typeof namespace.default === 'function' ||
      isAsyncFunction(namespace.default)
    ) {
      result = await namespace.default(task.payload);
    } else {
      result = namespace.default;
    }
  }

  // Remove the global event listener after it's no longer needed
  process.off('unhandledRejection', onUnhandledRejection);
  if (unhandledPromise) {
    return Promise.reject(new Error('An unhandled promise rejection occurred'));
  }

  return result;
}
