/**
 * @license
 * Copyright 2022 Dynatrace LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import type { JestEnvironment } from '@jest/environment';
import { LegacyFakeTimers, ModernFakeTimers } from '@jest/fake-timers';
import type { Config, Global } from '@jest/types';
import { ModuleMocker } from 'jest-mock';
import { installCommonGlobals } from 'jest-util';
import * as vm from 'vm';
import type { ContextConfiguration } from '../src/context';
import { createContext } from '../src/context';

export type Timer = {
  id: number;
  ref: () => Timer;
  unref: () => Timer;
};

/** Jest runtime simulator environment */
export default class RuntimeSimulatorEnvironment implements JestEnvironment {
  private readonly config: Config.ProjectConfig;
  private context: vm.Context | null;

  fakeTimers: LegacyFakeTimers<Timer> | null;
  fakeTimersModern: ModernFakeTimers | null;

  global: Global.Global;
  moduleMocker: ModuleMocker | null;

  /** Create a new test environment for the runtime simulator */
  constructor(config: Config.ProjectConfig) {
    this.config = config;

    // Note that `createContext` returns a function for reporting unhandled promise rejections;
    // however, Jest already provides its own handler for unhandled rejections that probably makes
    // more sense in a testing environment, so we can ignore it here
    this.context = createContext(
      process?.cwd() ?? '/',
      (globalThis.urlPathname as string) ?? '/api/test',
      this.config.testEnvironmentOptions! as unknown as ContextConfiguration,
    ).vmContext;

    const global = (this.global = vm.runInContext('this', this.context));
    global.global = global;
    global.clearInterval = clearInterval;
    global.clearTimeout = clearTimeout;
    global.setInterval = setInterval;
    global.setTimeout = setTimeout;
    global.Buffer = Buffer;

    installCommonGlobals(global, this.config.globals);

    this.moduleMocker = new ModuleMocker(global);

    const timerIdToRef = (id: number) => ({
      id,
      ref() {
        return this;
      },
      unref() {
        return this;
      },
    });
    const timerRefToId = (timer: Timer): number | undefined =>
      (timer && timer.id) || undefined;
    this.fakeTimers = new LegacyFakeTimers({
      config: this.config,
      global,
      moduleMocker: this.moduleMocker!,
      timerConfig: { idToRef: timerIdToRef, refToId: timerRefToId },
    });
    this.fakeTimersModern = new ModernFakeTimers({
      config: this.config,
      global,
    });
  }

  /** @inheritDoc */
  async setup(): Promise<void> {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const global = this.global as any;

    Object.assign(global, {
      // eslint-disable-next-line @typescript-eslint/no-explicit-any
      ...(this.config.testEnvironmentOptions?.globals as any),
      global,
      console: global.console,
      setTimeout: global.setTimeout,
      setInterval: global.setInterval,
      clearTimeout: global.clearTimeout,
      clearInterval: global.clearInterval,
    });
  }

  /** @inheritDoc */
  async teardown(): Promise<void> {
    this.fakeTimers?.dispose();
    this.fakeTimersModern?.dispose();
    this.context = null;
    this.fakeTimers = null;
    this.fakeTimersModern = null;
  }

  /** @inheritDoc */
  getVmContext(): vm.Context | null {
    return this.context;
  }
}
